"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.failProxy = void 0;
const assert_1 = require("assert");
const client_utils_1 = require("@fluid-internal/client-utils");
const test_pairwise_generator_1 = require("@fluid-private/test-pairwise-generator");
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../blobManager/index.js");
const failProxy = (handler = {}) => {
    const proxy = new Proxy(handler, {
        get: (t, p, r) => {
            if (p === "then") {
                return undefined;
            }
            if (handler !== undefined && p in handler) {
                return Reflect.get(t, p, r);
            }
            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            return (0, exports.failProxy)();
        },
    });
    return proxy;
};
exports.failProxy = failProxy;
function createBlobManager(overrides) {
    return new index_js_1.BlobManager((0, exports.failProxy)({
        // defaults, these can still be overridden below
        runtime: (0, exports.failProxy)({ baseLogger: (0, internal_2.createChildLogger)() }),
        snapshot: {},
        stashedBlobs: undefined,
        // overrides
        ...overrides,
    }));
}
const olderThanTTL = {
    blob: "olderThanTTL",
    minTTLInSeconds: 100,
    uploadTime: Date.now() - 100 * 1000,
};
const withinTTLHalfLife = {
    blob: "withinTTLHalfLife",
    storageId: "withinTTLHalfLife",
    minTTLInSeconds: 100,
    uploadTime: Date.now() - 25 * 1000,
};
const storageIdWithoutUploadTime = {
    blob: "storageIdWithoutUploadTime",
    storageId: "storageIdWithoutUploadTime",
};
describe("BlobManager.stashed", () => {
    it("No Pending Stashed Uploads", async () => {
        const blobManager = createBlobManager();
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        await blobManager.trackPendingStashedUploads();
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
    });
    it("Stashed blob", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                a: {
                    blob: "a",
                },
            },
            getStorage: () => (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Stashed blob with upload older than TTL", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                olderThanTTL,
            },
            getStorage: () => (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Stashed blob with upload within TTL half-life", async () => {
        const blobManager = createBlobManager({
            stashedBlobs: {
                withinTTLHalfLife,
            },
        });
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        await blobManager.trackPendingStashedUploads();
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
    });
    it("Stashed blob without upload time", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                storageIdWithoutUploadTime,
            },
            getStorage: () => (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Mixed stashed blobs", async () => {
        const letUploadsComplete = new internal_1.Deferred();
        const stashedBlobs = (0, test_pairwise_generator_1.generatePairwiseOptions)({
            acked: [undefined, true, false],
            blob: ["content"],
            minTTLInSeconds: [undefined, 100],
            storageId: [undefined, "id"],
            uploadTime: [Date.now() - 100 * 1000, Date.now() - 25 * 1000, undefined],
        }).reduce((pv, cv, i) => {
            if (cv.storageId) {
                cv.storageId += i.toString();
            }
            cv.blob = i.toString();
            pv[i] = cv;
            return pv;
        }, {});
        const blobManager = createBlobManager({
            stashedBlobs,
            getStorage: () => (0, exports.failProxy)({
                createBlob: async (b) => {
                    await letUploadsComplete.promise;
                    return { id: `id:${(0, client_utils_1.bufferToString)(b, "utf8")}` };
                },
            }),
        });
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        letUploadsComplete.resolve();
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.trackPendingStashedUploads(),
        ]);
        assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
});
//# sourceMappingURL=blobManager.stashed.spec.js.map