"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const benchmark_1 = require("@fluid-tools/benchmark");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../../gc/index.js");
const packageVersion_js_1 = require("../../packageVersion.js");
const gcUnitTestHelpers_js_1 = require("./gcUnitTestHelpers.js");
function createGarbageCollector(gcOptions) {
    const getNodeType = (nodePath) => {
        if (nodePath.split("/").length !== 2) {
            return index_js_1.GCNodeType.Other;
        }
        return index_js_1.GCNodeType.DataStore;
    };
    // The runtime to be passed to the garbage collector.
    const gcRuntime = {
        getGCData: async () => {
            return { gcNodes: {} };
        },
        updateUsedRoutes: (usedRoutes) => {
            return { totalNodeCount: 0, unusedNodeCount: 0 };
        },
        deleteSweepReadyNodes: (sweepReadyRoutes) => {
            return [];
        },
        updateTombstonedRoutes: (tombstoneRoutes) => { },
        getNodeType,
        getCurrentReferenceTimestampMs: () => Date.now(),
        closeFn: () => { },
    };
    return index_js_1.GarbageCollector.create({
        runtime: gcRuntime,
        gcOptions,
        baseSnapshot: undefined,
        baseLogger: (0, internal_1.createChildLogger)({}),
        existing: false,
        metadata: undefined,
        createContainerMetadata: {
            createContainerRuntimeVersion: packageVersion_js_1.pkgVersion,
            createContainerTimestamp: Date.now(),
        },
        isSummarizerClient: true,
        readAndParseBlob: gcUnitTestHelpers_js_1.parseNothing,
        getNodePackagePath: async (nodeId) => ["gcBenchmarkTestPkg"],
        getLastSummaryTimestampMs: () => Date.now(),
        submitMessage: (message) => { },
    });
}
describe("GC benchmark tests", () => {
    describe("initializeOrUpdateGCState", () => {
        const inactiveTimeoutMs = 100;
        let garbageCollector;
        /**
         * Sets up the garbage collector to have unreferenced nodes that will be updated
         * in the benchmark tests.
         */
        const setup = async (unrefNodeCount) => {
            garbageCollector = createGarbageCollector({ inactiveTimeoutMs });
            const currentTime = Date.now();
            // Set the unreferenced timestamp to older than inactive timeout so that the nodes start
            // as inactive.
            const unreferencedTimestampMs = currentTime - (inactiveTimeoutMs + 10);
            // Generate a base snapshot that has `unrefNodeCount` number of unreferenced nodes
            // and all are inactive.
            const gcState = { gcNodes: {} };
            for (let i = 0; i < unrefNodeCount; i++) {
                gcState.gcNodes[`node${i}`] = {
                    outboundRoutes: [],
                    unreferencedTimestampMs,
                };
            }
            const baseSnapshot = {
                gcState,
                tombstones: [],
                deletedNodes: [],
            };
            // Override the base snapshot promise to return the above snapshot.
            garbageCollector.baseSnapshotDataP = Promise.resolve(baseSnapshot);
            // Call initializeOrUpdateGCState once so that it initializes the GC state from the
            // base snapshot. The calls to initializeOrUpdateGCState after this one happens on
            // every connection and that's the one we are benchmarking
            await garbageCollector.initializeOrUpdateGCState();
        };
        const cleanup = () => {
            garbageCollector?.dispose();
        };
        /**
         * These tests benchmark how long it takes for the initializeOrUpdateGCState function to run
         * when there are a high number of unreferenced nodes. Since this results in clearing and
         * setting of timers, it takes a long time and can cause slow downs.
         */
        (0, benchmark_1.benchmark)({
            title: "5000 unref nodes",
            type: benchmark_1.BenchmarkType.Measurement,
            before: async () => setup(5000 /* unrefNodeCount */),
            benchmarkFnAsync: async () => {
                await garbageCollector.initializeOrUpdateGCState();
            },
            after: cleanup,
        });
        (0, benchmark_1.benchmark)({
            title: "15000 unref nodes",
            type: benchmark_1.BenchmarkType.Measurement,
            before: async () => setup(15000 /* unrefNodeCount */),
            benchmarkFnAsync: async () => {
                await garbageCollector.initializeOrUpdateGCState();
            },
            after: cleanup,
        });
        (0, benchmark_1.benchmark)({
            title: "30000 unref nodes",
            type: benchmark_1.BenchmarkType.Measurement,
            before: async () => setup(30000 /* unrefNodeCount */),
            benchmarkFnAsync: async () => {
                await garbageCollector.initializeOrUpdateGCState();
            },
            after: cleanup,
        });
    });
});
//# sourceMappingURL=gc.perf.spec.js.map