"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const test_pairwise_generator_1 = require("@fluid-private/test-pairwise-generator");
const internal_1 = require("@fluidframework/driver-definitions/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../../index.js");
const index_js_2 = require("../../opLifecycle/index.js");
describe("RemoteMessageProcessor", () => {
    function getMessageProcessor() {
        const logger = new internal_2.MockLogger();
        return new index_js_2.RemoteMessageProcessor(new index_js_2.OpSplitter([], undefined, 1, 1, logger), new index_js_2.OpDecompressor(logger), new index_js_2.OpGroupingManager({
            groupedBatchingEnabled: true,
            opCountThreshold: Infinity,
            reentrantBatchGroupingEnabled: false,
        }, logger));
    }
    function getOutboundMessage(value, batchMetadata) {
        return {
            metadata: batchMetadata === undefined
                ? undefined
                : {
                    batch: batchMetadata,
                },
            referenceSequenceNumber: Infinity,
            contents: JSON.stringify({
                contents: {
                    key: value,
                },
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
            }),
        };
    }
    function getProcessedMessage(value, seqNum, clientSeqNum, batchMetadata) {
        // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
        return {
            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
            metadata: batchMetadata === undefined
                ? undefined
                : {
                    batch: batchMetadata,
                },
            compression: undefined,
            sequenceNumber: seqNum,
            clientSequenceNumber: clientSeqNum,
            referenceSequenceNumber: Infinity,
            contents: {
                key: value,
            },
        };
    }
    const messageGenerationOptions = (0, test_pairwise_generator_1.generatePairwiseOptions)({
        compressionAndChunking: [
            { compression: false, chunking: false },
            { compression: true, chunking: false },
            { compression: true, chunking: true },
        ],
        grouping: [true, false],
    });
    messageGenerationOptions.forEach((option) => {
        it(`Correctly processes incoming messages: compression [${option.compressionAndChunking.compression}] chunking [${option.compressionAndChunking.chunking}] grouping [${option.grouping}]`, () => {
            let batch = {
                contentSizeInBytes: 1,
                referenceSequenceNumber: Infinity,
                messages: [
                    getOutboundMessage("a", true),
                    getOutboundMessage("b"),
                    getOutboundMessage("c"),
                    getOutboundMessage("d"),
                    getOutboundMessage("e", false),
                ],
            };
            const mockLogger = new internal_2.MockLogger();
            if (option.grouping) {
                const groupingManager = new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: true,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger);
                batch = groupingManager.groupBatch(batch);
            }
            let leadingChunkCount = 0;
            const outboundMessages = [];
            if (option.compressionAndChunking.compression) {
                const compressor = new index_js_2.OpCompressor(mockLogger);
                batch = compressor.compressBatch(batch);
                if (option.compressionAndChunking.chunking) {
                    const splitter = new index_js_2.OpSplitter([], (messages, refSeqNum) => {
                        ++leadingChunkCount;
                        outboundMessages.push(...messages);
                        return 0;
                    }, 2, Infinity, mockLogger);
                    batch = splitter.splitFirstBatchMessage(batch);
                }
            }
            let startSeqNum = outboundMessages.length + 1;
            outboundMessages.push(...batch.messages);
            const messageProcessor = getMessageProcessor();
            const actual = [];
            let seqNum = 1;
            let actualBatchStartCsn;
            let emptyProcessResultCount = 0;
            for (const message of outboundMessages) {
                // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
                const inboundMessage = {
                    type: internal_1.MessageType.Operation,
                    contents: message.contents,
                    metadata: message.metadata,
                    compression: message.compression,
                    sequenceNumber: seqNum,
                    clientSequenceNumber: seqNum++,
                    referenceSequenceNumber: message.referenceSequenceNumber,
                };
                (0, index_js_2.ensureContentsDeserialized)(inboundMessage, true, () => { });
                const processResult = messageProcessor.process(inboundMessage);
                // It'll be undefined for the first n-1 chunks if chunking is enabled
                if (processResult === undefined) {
                    ++emptyProcessResultCount;
                    continue;
                }
                actual.push(...processResult.messages);
                if (actualBatchStartCsn === undefined) {
                    actualBatchStartCsn = processResult.batchStartCsn;
                }
                else {
                    (0, assert_1.strict)(actualBatchStartCsn === processResult.batchStartCsn, "batchStartCsn shouldn't change while processing a single batch");
                }
            }
            assert_1.strict.equal(emptyProcessResultCount, leadingChunkCount, "expected empty result to be 1-1 with leading chunks");
            const expected = option.grouping
                ? [
                    getProcessedMessage("a", startSeqNum, 1, true),
                    getProcessedMessage("b", startSeqNum, 2),
                    getProcessedMessage("c", startSeqNum, 3),
                    getProcessedMessage("d", startSeqNum, 4),
                    getProcessedMessage("e", startSeqNum, 5, false),
                ]
                : [
                    getProcessedMessage("a", startSeqNum, startSeqNum++, true),
                    getProcessedMessage("b", startSeqNum, startSeqNum++),
                    getProcessedMessage("c", startSeqNum, startSeqNum++),
                    getProcessedMessage("d", startSeqNum, startSeqNum++),
                    getProcessedMessage("e", startSeqNum, startSeqNum, false),
                ];
            assert_1.strict.deepStrictEqual(actual, expected, "unexpected output");
            assert_1.strict.equal(actualBatchStartCsn, leadingChunkCount + 1, "unexpected batchStartCsn");
        });
    });
    it("Processes legacy string-content message", () => {
        const messageProcessor = getMessageProcessor();
        const contents = {
            contents: { key: "value" },
            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
        };
        const message = {
            contents: JSON.stringify(contents),
            clientId: "clientId",
            type: internal_1.MessageType.Operation,
            metadata: { meta: "data" },
        };
        const documentMessage = message;
        (0, index_js_2.ensureContentsDeserialized)(documentMessage, true, () => { });
        const processResult = messageProcessor.process(documentMessage)?.messages ?? [];
        assert_1.strict.strictEqual(processResult.length, 1, "only expected a single processed message");
        const result = processResult[0];
        assert_1.strict.deepStrictEqual(result.contents, contents.contents);
        assert_1.strict.deepStrictEqual(result.type, contents.type);
    });
    it("Don't unpack non-datastore messages", () => {
        const messageProcessor = getMessageProcessor();
        const message = {
            contents: { key: "value" },
            clientId: "clientId",
            type: internal_1.MessageType.Summarize,
            metadata: { meta: "data" },
        };
        const documentMessage = message;
        const processResult = messageProcessor.process(documentMessage)?.messages ?? [];
        assert_1.strict.strictEqual(processResult.length, 1, "only expected a single processed message");
        const result = processResult[0];
        assert_1.strict.deepStrictEqual(result.contents, message.contents);
        assert_1.strict.deepStrictEqual(result.type, message.type);
    });
    it("Processing groupedBatch works as expected", () => {
        const groupedBatch = {
            type: internal_1.MessageType.Operation,
            sequenceNumber: 10,
            clientSequenceNumber: 12,
            contents: {
                type: index_js_2.OpGroupingManager.groupedBatchOp,
                contents: [
                    {
                        contents: {
                            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                            contents: {
                                contents: "a",
                            },
                        },
                    },
                    {
                        contents: {
                            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                            contents: {
                                contents: "b",
                            },
                        },
                    },
                ],
            },
        };
        const messageProcessor = getMessageProcessor();
        const result = messageProcessor.process(groupedBatch);
        const expected = [
            {
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                sequenceNumber: 10,
                clientSequenceNumber: 1,
                compression: undefined,
                metadata: undefined,
                contents: {
                    contents: "a",
                },
            },
            {
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                sequenceNumber: 10,
                clientSequenceNumber: 2,
                compression: undefined,
                metadata: undefined,
                contents: {
                    contents: "b",
                },
            },
        ];
        assert_1.strict.deepStrictEqual(result, { messages: expected, batchStartCsn: 12 }, "unexpected processing of groupedBatch");
    });
});
//# sourceMappingURL=remoteMessageProcessor.spec.js.map