/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ICriticalContainerError } from "@fluidframework/container-definitions";
import { IDisposable } from "@fluidframework/core-interfaces";
import { ITelemetryLoggerExt } from "@fluidframework/telemetry-utils/internal";
import { InboundSequencedContainerRuntimeMessage } from "./messageTypes.js";
import { BatchId, BatchMessage } from "./opLifecycle/index.js";
/**
 * This represents a message that has been submitted and is added to the pending queue when `submit` is called on the
 * ContainerRuntime. This message has either not been ack'd by the server or has not been submitted to the server yet.
 *
 * @remarks This is the current serialization format for pending local state when a Container is serialized.
 */
export interface IPendingMessage {
    type: "message";
    referenceSequenceNumber: number;
    content: string;
    localOpMetadata: unknown;
    opMetadata: Record<string, unknown> | undefined;
    sequenceNumber?: number;
    /** Info needed to compute the batchId on reconnect */
    batchIdContext: {
        /** The Batch's original clientId, from when it was first flushed to be submitted */
        clientId: string;
        /**
         * The Batch's original clientSequenceNumber, from when it was first flushed to be submitted
         *	@remarks A negative value means it was not yet submitted when queued here (e.g. disconnected right before flush fired)
         */
        batchStartCsn: number;
    };
}
export interface IPendingLocalState {
    /**
     * list of pending states, including ops and batch information
     */
    pendingStates: IPendingMessage[];
}
/** Info needed to replay/resubmit a pending message */
export type PendingMessageResubmitData = Pick<IPendingMessage, "content" | "localOpMetadata" | "opMetadata">;
export interface IRuntimeStateHandler {
    connected(): boolean;
    clientId(): string | undefined;
    close(error?: ICriticalContainerError): void;
    applyStashedOp(content: string): Promise<unknown>;
    reSubmitBatch(batch: PendingMessageResubmitData[], batchId: BatchId): void;
    isActiveConnection: () => boolean;
    isAttached: () => boolean;
}
/**
 * PendingStateManager is responsible for maintaining the messages that have not been sent or have not yet been
 * acknowledged by the server. It also maintains the batch information for both automatically and manually flushed
 * batches along with the messages.
 * When the Container reconnects, it replays the pending states, which includes manual flushing
 * of messages and triggering resubmission of unacked ops.
 *
 * It verifies that all the ops are acked, are received in the right order and batch information is correct.
 */
export declare class PendingStateManager implements IDisposable {
    private readonly stateHandler;
    private readonly logger;
    /** Messages that will need to be resubmitted if not ack'd before the next reconnection */
    private readonly pendingMessages;
    /** Messages stashed from a previous container, now being rehydrated. Need to be resubmitted. */
    private readonly initialMessages;
    /**
     * Sequenced local ops that are saved when stashing since pending ops may depend on them
     */
    private savedOps;
    /** Used to stand in for batchStartCsn for messages that weren't submitted (so no CSN) */
    private negativeCounter;
    private readonly disposeOnce;
    private isProcessingBatch;
    private pendingBatchBeginMessage;
    /** Used to ensure we don't replay ops on the same connection twice */
    private clientIdFromLastReplay;
    /**
     * The pending messages count. Includes `pendingMessages` and `initialMessages` to keep in sync with
     * 'hasPendingMessages'.
     */
    get pendingMessagesCount(): number;
    /**
     * The minimumPendingMessageSequenceNumber is the minimum of the first pending message and the first initial message.
     *
     * We need this so that we can properly keep local data and maintain the correct sequence window.
     */
    get minimumPendingMessageSequenceNumber(): number | undefined;
    /**
     * Called to check if there are any pending messages in the pending message queue.
     * @returns A boolean indicating whether there are messages or not.
     */
    hasPendingMessages(): boolean;
    getLocalState(snapshotSequenceNumber?: number): IPendingLocalState;
    constructor(stateHandler: IRuntimeStateHandler, stashedLocalState: IPendingLocalState | undefined, logger: ITelemetryLoggerExt);
    get disposed(): boolean;
    readonly dispose: () => void;
    /**
     * The given batch has been flushed, and needs to be tracked locally until the corresponding
     * acks are processed, to ensure it is successfully sent.
     * @param batch - The batch that was flushed
     * @param clientSequenceNumber - The CSN of the first message in the batch,
     * or undefined if the batch was not yet sent (e.g. by the time we flushed we lost the connection)
     */
    onFlushBatch(batch: BatchMessage[], clientSequenceNumber: number | undefined): void;
    /**
     * Applies stashed ops at their reference sequence number so they are ready to be ACKed or resubmitted
     * @param seqNum - Sequence number at which to apply ops. Will apply all ops if seqNum is undefined.
     */
    applyStashedOpsAt(seqNum?: number): Promise<void>;
    /**
     * Processes a local message once its ack'd by the server. It verifies that there was no data corruption and that
     * the batch information was preserved for batch messages.
     * @param message - The message that got ack'd and needs to be processed.
     * @param batchStartCsn - The clientSequenceNumber of the start of this message's batch (assigned during submit)
     * (not to be confused with message.clientSequenceNumber - the overwritten value in case of grouped batching)
     */
    processPendingLocalMessage(message: InboundSequencedContainerRuntimeMessage, batchStartCsn: number): unknown;
    /**
     * This message could be the first message in batch. If so, set batch state marking the beginning of a batch.
     * @param message - The message that is being processed.
     * @param batchStartCsn - The clientSequenceNumber of the start of this message's batch (assigned during submit)
     * @param pendingMessage - The corresponding pendingMessage.
     */
    private maybeProcessBatchBegin;
    /**
     * This message could be the last message in batch. If so, clear batch state since the batch is complete.
     * @param message - The message that is being processed.
     */
    private maybeProcessBatchEnd;
    /**
     * Called when the Container's connection state changes. If the Container gets connected, it replays all the pending
     * states in its queue. This includes triggering resubmission of unacked ops.
     * ! Note: successfully resubmitting an op that has been successfully sequenced is not possible due to checks in the ConnectionStateHandler (Loader layer)
     */
    replayPendingStates(): void;
}
//# sourceMappingURL=pendingStateManager.d.ts.map