"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PendingStateManager = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const double_ended_queue_1 = __importDefault(require("double-ended-queue"));
const uuid_1 = require("uuid");
const metadata_js_1 = require("./metadata.js");
const index_js_1 = require("./opLifecycle/index.js");
const packageVersion_js_1 = require("./packageVersion.js");
function buildPendingMessageContent(
// AnyComboFromUnion is needed need to gain access to compatDetails that
// is only defined for some cases.
message) {
    // IMPORTANT: Order matters here, this must match the order of the properties used
    // when submitting the message.
    const { type, contents, compatDetails } = message;
    // Any properties that are not defined, won't be emitted by stringify.
    return JSON.stringify({ type, contents, compatDetails });
}
function withoutLocalOpMetadata(message) {
    return {
        ...message,
        localOpMetadata: undefined,
    };
}
/**
 * PendingStateManager is responsible for maintaining the messages that have not been sent or have not yet been
 * acknowledged by the server. It also maintains the batch information for both automatically and manually flushed
 * batches along with the messages.
 * When the Container reconnects, it replays the pending states, which includes manual flushing
 * of messages and triggering resubmission of unacked ops.
 *
 * It verifies that all the ops are acked, are received in the right order and batch information is correct.
 */
class PendingStateManager {
    /**
     * The pending messages count. Includes `pendingMessages` and `initialMessages` to keep in sync with
     * 'hasPendingMessages'.
     */
    get pendingMessagesCount() {
        return this.pendingMessages.length + this.initialMessages.length;
    }
    /**
     * The minimumPendingMessageSequenceNumber is the minimum of the first pending message and the first initial message.
     *
     * We need this so that we can properly keep local data and maintain the correct sequence window.
     */
    get minimumPendingMessageSequenceNumber() {
        return this.pendingMessages.peekFront()?.referenceSequenceNumber;
    }
    /**
     * Called to check if there are any pending messages in the pending message queue.
     * @returns A boolean indicating whether there are messages or not.
     */
    hasPendingMessages() {
        return this.pendingMessagesCount !== 0;
    }
    getLocalState(snapshotSequenceNumber) {
        (0, internal_1.assert)(this.initialMessages.isEmpty(), 0x2e9 /* "Must call getLocalState() after applying initial states" */);
        // Using snapshot sequence number to filter ops older than our latest snapshot.
        // Such ops should not be declared in pending/stashed state. Snapshot seq num will not
        // be available when the container is not attached. Therefore, no filtering is needed.
        const newSavedOps = [...this.savedOps].filter((message) => {
            (0, internal_1.assert)(message.sequenceNumber !== undefined, 0x97c /* saved op should already have a sequence number */);
            return message.sequenceNumber > (snapshotSequenceNumber ?? 0);
        });
        this.pendingMessages.toArray().forEach((message) => {
            if (snapshotSequenceNumber !== undefined &&
                message.referenceSequenceNumber < snapshotSequenceNumber) {
                throw new internal_2.LoggingError("trying to stash ops older than our latest snapshot");
            }
        });
        return {
            pendingStates: [
                ...newSavedOps,
                ...this.pendingMessages.toArray().map(withoutLocalOpMetadata),
            ],
        };
    }
    constructor(stateHandler, stashedLocalState, logger) {
        this.stateHandler = stateHandler;
        this.logger = logger;
        /** Messages that will need to be resubmitted if not ack'd before the next reconnection */
        this.pendingMessages = new double_ended_queue_1.default();
        /** Messages stashed from a previous container, now being rehydrated. Need to be resubmitted. */
        this.initialMessages = new double_ended_queue_1.default();
        /**
         * Sequenced local ops that are saved when stashing since pending ops may depend on them
         */
        this.savedOps = [];
        /** Used to stand in for batchStartCsn for messages that weren't submitted (so no CSN) */
        this.negativeCounter = -1;
        this.disposeOnce = new internal_1.Lazy(() => {
            this.initialMessages.clear();
            this.pendingMessages.clear();
        });
        // Indicates whether we are processing a batch.
        this.isProcessingBatch = false;
        this.dispose = () => this.disposeOnce.value;
        if (stashedLocalState?.pendingStates) {
            this.initialMessages.push(...stashedLocalState.pendingStates);
        }
    }
    get disposed() {
        return this.disposeOnce.evaluated;
    }
    /**
     * The given batch has been flushed, and needs to be tracked locally until the corresponding
     * acks are processed, to ensure it is successfully sent.
     * @param batch - The batch that was flushed
     * @param clientSequenceNumber - The CSN of the first message in the batch,
     * or undefined if the batch was not yet sent (e.g. by the time we flushed we lost the connection)
     */
    onFlushBatch(batch, clientSequenceNumber) {
        // If we're connected this is the client of the current connection,
        // otherwise it's the clientId that just disconnected
        // It's only undefined if we've NEVER connected. This is a tight corner case and we can
        // simply make up a unique ID in this case.
        const clientId = this.stateHandler.clientId() ?? (0, uuid_1.v4)();
        // If the batch was not yet sent, we need to assign a unique batchStartCsn
        // Use a negative number to distinguish these from real CSNs
        const batchStartCsn = clientSequenceNumber ?? this.negativeCounter--;
        for (const message of batch) {
            const { contents: content = "", referenceSequenceNumber, localOpMetadata, metadata: opMetadata, } = message;
            const pendingMessage = {
                type: "message",
                referenceSequenceNumber,
                content,
                localOpMetadata,
                opMetadata,
                // Note: We only need this on the first message.
                batchIdContext: { clientId, batchStartCsn },
            };
            this.pendingMessages.push(pendingMessage);
        }
    }
    /**
     * Applies stashed ops at their reference sequence number so they are ready to be ACKed or resubmitted
     * @param seqNum - Sequence number at which to apply ops. Will apply all ops if seqNum is undefined.
     */
    async applyStashedOpsAt(seqNum) {
        // apply stashed ops at sequence number
        while (!this.initialMessages.isEmpty()) {
            if (seqNum !== undefined) {
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                const peekMessage = this.initialMessages.peekFront();
                if (peekMessage.referenceSequenceNumber > seqNum) {
                    break; // nothing left to do at this sequence number
                }
                if (peekMessage.referenceSequenceNumber < seqNum) {
                    throw new Error("loaded from snapshot too recent to apply stashed ops");
                }
            }
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            const nextMessage = this.initialMessages.shift();
            try {
                // applyStashedOp will cause the DDS to behave as if it has sent the op but not actually send it
                const localOpMetadata = await this.stateHandler.applyStashedOp(nextMessage.content);
                if (!this.stateHandler.isAttached()) {
                    if (localOpMetadata !== undefined) {
                        throw new Error("Local Op Metadata must be undefined when not attached");
                    }
                }
                else {
                    nextMessage.localOpMetadata = localOpMetadata;
                    // then we push onto pendingMessages which will cause PendingStateManager to resubmit when we connect
                    patchBatchIdContext(nextMessage); // Back compat
                    this.pendingMessages.push(nextMessage);
                }
            }
            catch (error) {
                throw internal_2.DataProcessingError.wrapIfUnrecognized(error, "applyStashedOp", nextMessage);
            }
        }
    }
    /**
     * Processes a local message once its ack'd by the server. It verifies that there was no data corruption and that
     * the batch information was preserved for batch messages.
     * @param message - The message that got ack'd and needs to be processed.
     * @param batchStartCsn - The clientSequenceNumber of the start of this message's batch (assigned during submit)
     * (not to be confused with message.clientSequenceNumber - the overwritten value in case of grouped batching)
     */
    processPendingLocalMessage(message, batchStartCsn) {
        // Get the next message from the pending queue. Verify a message exists.
        const pendingMessage = this.pendingMessages.peekFront();
        (0, internal_1.assert)(pendingMessage !== undefined, 0x169 /* "No pending message found for this remote message" */);
        // This may be the start of a batch.
        this.maybeProcessBatchBegin(message, batchStartCsn, pendingMessage);
        pendingMessage.sequenceNumber = message.sequenceNumber;
        this.savedOps.push(withoutLocalOpMetadata(pendingMessage));
        this.pendingMessages.shift();
        const messageContent = buildPendingMessageContent(message);
        // Stringified content should match
        if (pendingMessage.content !== messageContent) {
            this.stateHandler.close(internal_2.DataProcessingError.create("pending local message content mismatch", "unexpectedAckReceived", message, {
                expectedMessageType: JSON.parse(pendingMessage.content).type,
            }));
            return;
        }
        // Post-processing part - If we are processing a batch then this could be the last message in the batch.
        this.maybeProcessBatchEnd(message);
        return pendingMessage.localOpMetadata;
    }
    /**
     * This message could be the first message in batch. If so, set batch state marking the beginning of a batch.
     * @param message - The message that is being processed.
     * @param batchStartCsn - The clientSequenceNumber of the start of this message's batch (assigned during submit)
     * @param pendingMessage - The corresponding pendingMessage.
     */
    maybeProcessBatchBegin(message, batchStartCsn, pendingMessage) {
        if (!this.isProcessingBatch) {
            // Expecting the start of a batch (maybe single-message).
            if (pendingMessage.batchIdContext.batchStartCsn !== batchStartCsn) {
                this.logger?.sendErrorEvent({
                    eventName: "BatchClientSequenceNumberMismatch",
                    details: {
                        processingBatch: !!this.pendingBatchBeginMessage,
                        pendingBatchCsn: pendingMessage.batchIdContext.batchStartCsn,
                        batchStartCsn,
                        messageBatchMetadata: message.metadata?.batch,
                        pendingMessageBatchMetadata: pendingMessage.opMetadata?.batch,
                    },
                    messageDetails: (0, internal_2.extractSafePropertiesFromMessage)(message),
                });
            }
        }
        // This message is the first in a batch if the "batch" property on the metadata is set to true
        if (message.metadata?.batch) {
            // We should not already be processing a batch and there should be no pending batch begin message.
            (0, internal_1.assert)(!this.isProcessingBatch && this.pendingBatchBeginMessage === undefined, 0x16b /* "The pending batch state indicates we are already processing a batch" */);
            // Set the pending batch state indicating we have started processing a batch.
            this.pendingBatchBeginMessage = message;
            this.isProcessingBatch = true;
        }
    }
    /**
     * This message could be the last message in batch. If so, clear batch state since the batch is complete.
     * @param message - The message that is being processed.
     */
    maybeProcessBatchEnd(message) {
        if (!this.isProcessingBatch) {
            return;
        }
        // There should be a pending batch begin message.
        (0, internal_1.assert)(this.pendingBatchBeginMessage !== undefined, 0x16d /* "There is no pending batch begin message" */);
        const batchEndMetadata = message.metadata?.batch;
        if (this.pendingMessages.isEmpty() || batchEndMetadata === false) {
            // Get the batch begin metadata from the first message in the batch.
            const batchBeginMetadata = this.pendingBatchBeginMessage.metadata?.batch;
            // There could be just a single message in the batch. If so, it should not have any batch metadata. If there
            // are multiple messages in the batch, verify that we got the correct batch begin and end metadata.
            if (this.pendingBatchBeginMessage === message) {
                (0, internal_1.assert)(batchBeginMetadata === undefined, 0x16e /* "Batch with single message should not have batch metadata" */);
            }
            else {
                if (batchBeginMetadata !== true || batchEndMetadata !== false) {
                    this.stateHandler.close(internal_2.DataProcessingError.create("Pending batch inconsistency", // Formerly known as asserts 0x16f and 0x170
                    "processPendingLocalMessage", message, {
                        runtimeVersion: packageVersion_js_1.pkgVersion,
                        batchClientId: 
                        // eslint-disable-next-line @typescript-eslint/prefer-nullish-coalescing
                        this.pendingBatchBeginMessage.clientId === null
                            ? "null"
                            : this.pendingBatchBeginMessage.clientId,
                        clientId: this.stateHandler.clientId(),
                        hasBatchStart: batchBeginMetadata === true,
                        hasBatchEnd: batchEndMetadata === false,
                        messageType: message.type,
                        pendingMessagesCount: this.pendingMessagesCount,
                    }));
                }
            }
            // Clear the pending batch state now that we have processed the entire batch.
            this.pendingBatchBeginMessage = undefined;
            this.isProcessingBatch = false;
        }
    }
    /**
     * Called when the Container's connection state changes. If the Container gets connected, it replays all the pending
     * states in its queue. This includes triggering resubmission of unacked ops.
     * ! Note: successfully resubmitting an op that has been successfully sequenced is not possible due to checks in the ConnectionStateHandler (Loader layer)
     */
    replayPendingStates() {
        (0, internal_1.assert)(this.stateHandler.connected(), 0x172 /* "The connection state is not consistent with the runtime" */);
        // This assert suggests we are about to send same ops twice, which will result in data loss.
        (0, internal_1.assert)(this.clientIdFromLastReplay !== this.stateHandler.clientId(), 0x173 /* "replayPendingStates called twice for same clientId!" */);
        this.clientIdFromLastReplay = this.stateHandler.clientId();
        (0, internal_1.assert)(this.initialMessages.isEmpty(), 0x174 /* "initial states should be empty before replaying pending" */);
        const initialPendingMessagesCount = this.pendingMessages.length;
        let remainingPendingMessagesCount = this.pendingMessages.length;
        // Process exactly `pendingMessagesCount` items in the queue as it represents the number of messages that were
        // pending when we connected. This is important because the `reSubmitFn` might add more items in the queue
        // which must not be replayed.
        while (remainingPendingMessagesCount > 0) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            let pendingMessage = this.pendingMessages.shift();
            remainingPendingMessagesCount--;
            const batchMetadataFlag = (0, metadata_js_1.asBatchMetadata)(pendingMessage.opMetadata)?.batch;
            (0, internal_1.assert)(batchMetadataFlag !== false, 0x41b /* We cannot process batches in chunks */);
            // The next message starts a batch (possibly single-message), and we'll need its batchId.
            // We'll find batchId on this message if it was previously generated.
            // Otherwise, generate it now - this is the first time resubmitting this batch.
            const batchId = (0, metadata_js_1.asBatchMetadata)(pendingMessage.opMetadata)?.batchId ??
                (0, index_js_1.generateBatchId)(pendingMessage.batchIdContext.clientId, pendingMessage.batchIdContext.batchStartCsn);
            /**
             * We must preserve the distinct batches on resubmit.
             * Note: It is not possible for the PendingStateManager to receive a partially acked batch. It will
             * either receive the whole batch ack or nothing at all.  @see ScheduleManager for how this works.
             */
            if (batchMetadataFlag === undefined) {
                // Single-message batch
                this.stateHandler.reSubmitBatch([
                    {
                        content: pendingMessage.content,
                        localOpMetadata: pendingMessage.localOpMetadata,
                        opMetadata: pendingMessage.opMetadata,
                    },
                ], batchId);
                continue;
            }
            // else: batchMetadataFlag === true  (It's a typical multi-message batch)
            (0, internal_1.assert)(remainingPendingMessagesCount > 0, 0x554 /* Last pending message cannot be a batch begin */);
            const batch = [];
            // check is >= because batch end may be last pending message
            while (remainingPendingMessagesCount >= 0) {
                batch.push({
                    content: pendingMessage.content,
                    localOpMetadata: pendingMessage.localOpMetadata,
                    opMetadata: pendingMessage.opMetadata,
                });
                if (pendingMessage.opMetadata?.batch === false) {
                    break;
                }
                (0, internal_1.assert)(remainingPendingMessagesCount > 0, 0x555 /* No batch end found */);
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                pendingMessage = this.pendingMessages.shift();
                remainingPendingMessagesCount--;
                (0, internal_1.assert)(pendingMessage.opMetadata?.batch !== true, 0x556 /* Batch start needs a corresponding batch end */);
            }
            this.stateHandler.reSubmitBatch(batch, batchId);
        }
        // pending ops should no longer depend on previous sequenced local ops after resubmit
        this.savedOps = [];
        // We replayPendingStates on read connections too - we expect these to get nack'd though, and to then reconnect
        // on a write connection and replay again. This filters out the replay that happens on the read connection so
        // we only see the replays on write connections (that have a chance to go through).
        if (this.stateHandler.isActiveConnection()) {
            this.logger?.sendTelemetryEvent({
                eventName: "PendingStatesReplayed",
                count: initialPendingMessagesCount,
                clientId: this.stateHandler.clientId(),
            });
        }
    }
}
exports.PendingStateManager = PendingStateManager;
/** For back-compat if trying to apply stashed ops that pre-date batchIdContext */
function patchBatchIdContext(message) {
    const batchIdContext = message.batchIdContext;
    if (batchIdContext === undefined) {
        // Using uuid guarantees uniqueness, retaining existing behavior
        message.batchIdContext = { clientId: (0, uuid_1.v4)(), batchStartCsn: -1 };
    }
}
//# sourceMappingURL=pendingStateManager.js.map